==> GotoBLAS-1.00.tar.gz: GotoBLAS/level1/others/memory.c <==
/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#undef  DEBUG
#undef  DEBUG_AFFINITY

#if (!defined(ARCH_X86) && !defined(ARCH_X86_64)) || \
     !defined(SMP_SERVER) || !defined(linux) || (MAX_CPU_NUMBER == 1)
#undef USE_CPU_AFFINITY
#undef USE_MEMORY_AFFINITY
#endif

#if defined(linux) && !defined(USE_CPU_AFFINITY) && defined(USE_MEMORY_AFFINITY)
#define USE_CPU_AFFINITY
#endif

#ifndef ALLOC_STATIC
#if defined( __WIN32__) || defined(__CYGWIN32__)
#define ALLOC_WINDOWS
#else
#define ALLOC_MMAP
#define ALLOC_MALLOC
#endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#ifdef __unix__
#include <pthread.h>
#include <sys/mman.h>
#endif
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#ifdef linux
#include <sys/sysinfo.h>
#include <sched.h>
#include <errno.h>
#include <linux/unistd.h>
#ifdef USE_MEMORY_AFFINITY
#include <linux/mempolicy.h>
#endif
#endif

#if defined( __WIN32__) || defined(__CYGWIN32__)
#include <windows.h>
#endif

#include "common.h"

#ifdef ALLOC_HUGETLB
#define SHM_HUGETLB 04000
#endif

#ifdef USE_CPU_AFFINITY
int sched_getaffinity(pid_t pid, unsigned int len, unsigned long *mask);
int sched_setaffinity(pid_t pid, unsigned int len, unsigned long *mask);

static inline int get_physical_pos(void){
  int eax, ebx, ecx, edx;
  int apicid;

  cpuid(1, &eax, &ebx, &ecx, &edx);
  apicid  = BITMASK(ebx, 24, 0xff);

  return apicid;
}
#endif

#ifdef USE_MEMORY_AFFINITY
_syscall6(int, mbind, void *, start, unsigned long, len, int, policy, unsigned \
long *, mask, unsigned long, maxnode, unsigned int, flag);

/* This function works only on Opteron/Prescott */
static inline int get_num_cores(void){

  int ncores;

#if   defined(CORE_PRESCOTT)
  int eax, ebx, ecx, edx;

  cpuid(4, &eax, &ebx, &ecx, &edx);
  ncores = BITMASK(eax, 26, 0x3f) + 1;
#elif defined(CORE_OPTERON)
  int eax, ebx, ecx, edx;

  cpuid(0x80000008, &eax, &ebx, &ecx, &edx);
  ncores = BITMASK(ecx,  0, 0xff) + 1;
#else
  ncores = 1;
#endif

  return ncores;
}
#endif

#ifndef FIXED_PAGESIZE
#define FIXED_PAGESIZE 4096
#endif

#define BITMASK(a, b, c) ((((a) >> (b)) & (c)))

#ifndef SMP

#define blas_cpu_number 1

#else

/* Global Parameter */
int blas_cpu_number = 0;

int blas_get_cpu_number(void){
  char *p;

  if (blas_cpu_number) return blas_cpu_number;

  p = getenv("GOTO_NUM_THREADS");

  if (p == (char *)NULL) p = getenv("OMP_NUM_THREADS");

  if (p) {
    blas_cpu_number = atoi(p);
#ifdef DEBUG
    printf( "Environment is defined : %3d\n", blas_cpu_number);
#endif
  } else {
#ifdef linux
    blas_cpu_number = get_nprocs();
#endif
  }

  if (blas_cpu_number <= 0) blas_cpu_number = DEFAULT_CPU_NUMBER;

  if (blas_cpu_number > MAX_CPU_NUMBER) blas_cpu_number = MAX_CPU_NUMBER;

#ifdef DEBUG
    printf( "Adjusted number of threads : %3d\n", blas_cpu_number);
#endif

  return blas_cpu_number;
}
#endif

#ifdef ALLOC_STATIC
extern double alloc_area[MAX_CPU_NUMBER * BUFFER_SIZE / sizeof(double)];
#else
#ifdef SEEK_ADDRESS
static unsigned long base_address      = 0UL;
#else
static unsigned long base_address      = BASE_ADDRESS;
#endif
static unsigned long alloc_area[MAX_CPU_NUMBER];
#endif

static int alloc_used[MAX_CPU_NUMBER];

#if   defined(USE_PTHREAD_LOCK)
static pthread_mutex_t    alloc_lock = PTHREAD_MUTEX_INITIALIZER;
#elif defined(USE_PTHREAD_SPINLOCK)
static pthread_spinlock_t alloc_lock = 0;
#else
static unsigned long      alloc_lock = 0UL;
#endif

#ifdef ALLOC_MMAP
static void *alloc_mmap(void *address){
  void *map_address;

  if (address){
    map_address = mmap(address, 
		       BUFFER_SIZE, 
		       MMAP_ACCESS, MMAP_POLICY | MAP_FIXED, -1, 0);
  } else {
    map_address = mmap(address, 
		       BUFFER_SIZE, 
		       MMAP_ACCESS, MMAP_POLICY, -1, 0);
  }

  return map_address;
}
#endif

#ifdef ALLOC_MALLOC
static void *alloc_malloc(void *address){
  void *map_address;

  map_address = (void *)malloc(BUFFER_SIZE + FIXED_PAGESIZE);

  /* Dangerous. Never release this area */
  return (void *)(((long)map_address + FIXED_PAGESIZE - 1) & ~(FIXED_PAGESIZE - 1));
}
#endif

#ifdef ALLOC_QALLOC
void *qalloc(int flags, size_t bytes);
#define QNONCACHE 0x1
#define QCOMMS    0x2
#define QFAST     0x4

static void *alloc_qalloc(void *address){
  void *map_address;

  map_address = (void *)qalloc(QCOMMS | QFAST, BUFFER_SIZE + FIXED_PAGESIZE);

  return (void *)(((long)map_address + FIXED_PAGESIZE - 1) & ~(FIXED_PAGESIZE - 1));
}
#endif

#ifdef ALLOC_WINDOWS
static void *alloc_windows(void *address){
  void *map_address;

  map_address  = VirtualAlloc(address,
			      BUFFER_SIZE,
			      MEM_RESERVE | MEM_COMMIT,
			      PAGE_READWRITE);
  return map_address;
}
#endif

#ifdef ALLOC_SHM
static void *alloc_shm(void *address){
  unsigned long map_address;
  int shmid;
  
  shmid = shmget(IPC_PRIVATE, BUFFER_SIZE,IPC_CREAT | 0600);
  
  map_address = (unsigned long)shmat(shmid, address, 0);

  if ((long)map_address != -1){
    shmctl(shmid, IPC_RMID, 0);
  }

  return (void *)map_address;
}
#endif

#ifdef ALLOC_HUGETLB
static void *alloc_hugetlb(void *address){
  unsigned long map_address = (unsigned long)-1;

#if defined(linux) || defined(_AIX)
  int shmid;
  
  shmid = shmget(IPC_PRIVATE, BUFFER_SIZE,
#ifdef linux
		 SHM_HUGETLB |
#endif
#ifdef _AIX
		 SHM_LGPAGE | SHM_PIN |
#endif
		 IPC_CREAT | SHM_R | SHM_W);
  
  if (shmid != -1) {
    map_address = (unsigned long)shmat(shmid, address, SHM_RND);
    
    if ((long)map_address != -1){
      shmctl(shmid, IPC_RMID, 0);
    }
  }
#endif

#ifdef __sparc__
  struct memcntl_mha mha;
  
  mha.mha_cmd = MHA_MAPSIZE_BSSBRK;
  mha.mha_flags = 0;
  mha.mha_pagesize = HUGE_PAGESIZE;
  memcntl(NULL, 0, MC_HAT_ADVISE, (char *)&mha, 0, 0);

  map_address = (unsigned long)memalign(HUGE_PAGESIZE, BUFFER_SIZE);
#endif

  return (void *)map_address;
}
#endif

#if defined(ARCH_X86) || defined(ARCH_X86_64) || defined(ARCH_IA64)
static int              initialized = 0;
#endif

void *blas_memory_alloc(int numproc){
  
  int position;
#ifndef ALLOC_STATIC
  unsigned long map_address;
#endif
#ifdef USE_CPU_AFFINITY
  unsigned long mask, rv;
#endif

  void *(*memoryalloc[])(void *address) = {
#ifdef ALLOC_HUGETLB
    alloc_hugetlb,
#endif
#ifdef ALLOC_SHM
    alloc_shm,
#endif
#ifdef ALLOC_MMAP
    alloc_mmap,
#endif
#ifdef ALLOC_QALLOC
    alloc_qalloc,
#endif
#ifdef ALLOC_WINDOWS
    alloc_windows,
#endif
#ifdef ALLOC_MALLOC
    alloc_malloc,
#endif
    NULL,
  };
  void *(**func)(void *address);

#ifdef SMP
  if (!blas_cpu_number) blas_get_cpu_number();
#endif

  LOCK_COMMAND(&alloc_lock);
  
#ifdef DEBUG
  printf("Alloc Start ...\n");
#endif

  /* main thread allocate to main CPU */
#if MAX_CPU_NUMBER == 1
  position = 0;
#else
  position = numproc;
#endif

#ifdef USE_CPU_AFFINITY
  position = get_physical_pos();
  if ((position >= MAX_CPU_NUMBER) || alloc_used[position])
    position = MAX_CPU_NUMBER - 1;
  while (alloc_used[position]) position--;
#else
  while (alloc_used[position]) position++;
#endif

  if ((position < 0) || (position >= MAX_CPU_NUMBER)) goto error;

#ifdef DEBUG
  printf("  Position -> %d\n", position);
#endif

#ifndef ALLOC_STATIC

  if (!alloc_area[position]) {
    do {
#ifdef DEBUG
      printf("Allocation Start : %lx\n", base_address);
#endif

#ifdef USE_CPU_AFFINITY
      rv = sched_getaffinity((pid_t)0, sizeof(mask), &mask);
      if ((rv == 0) && (mask & (1 << position))) {
	mask = (1 << position);
	rv = sched_setaffinity((pid_t)0, sizeof(mask), &mask);
      }
#endif

      map_address = (unsigned long)-1;
      func = &memoryalloc[0];

      while ((func != NULL) && (map_address == (unsigned long) -1)) {
	map_address = (unsigned long)(*func)((void *)base_address);
	func ++;
      }

#ifdef USE_MEMORY_AFFINITY
      if (map_address) {
	mask = (1 << (position >> get_num_cores()));
	rv = mbind((void *)map_address, BUFFER_SIZE, 
		   MPOL_BIND, &mask, sizeof(mask), MPOL_MF_STRICT);
      }
#endif

#ifdef DEBUG
      printf("  Success -> %08lx\n", map_address);
#endif
      if (((long) map_address) == -1) base_address = 0UL;
      if (base_address) base_address += BUFFER_SIZE;

    } while ((long)map_address == -1);

    alloc_area[position] = map_address;

#ifdef DEBUG
    printf("  Mapping Succeeded. %p(%d)\n", (void *)alloc_area[position], position);
#endif
  }
 
#endif

  alloc_used[position] = 1;

#ifdef DEBUG
  printf("Mapped   : %p  %3d\n\n",
	  (void *)alloc_area[position], position);
#endif  

#if defined(ARCH_X86) || defined(ARCH_X86_64) || defined(ARCH_IA64)
  if (initialized == 0) {
    blas_set_parameter();
    initialized = 1;
  }
#endif

  UNLOCK_COMMAND(&alloc_lock);
  
#ifndef ALLOC_STATIC
  return (void *)alloc_area[position];
#else
  return ((void *)alloc_area) + position * BUFFER_SIZE;
#endif

 error:
  UNLOCK_COMMAND(&alloc_lock);
  printf("BLAS : Bad memory allocation! Program is Terminated.\n");
  exit(1);
}

void blas_memory_free(void *free_area){

  int position;

  LOCK_COMMAND(&alloc_lock);

#ifdef DEBUG
  printf("Unmapped Start : %p ...\n", free_area);
#endif  

#ifndef ALLOC_STATIC
  position = 0;

  while (((unsigned long)alloc_area[position] != (unsigned long)free_area)
	 && (position < MAX_CPU_NUMBER)) position++;

  if ((unsigned long)alloc_area[position] != (unsigned long)free_area) goto error;
#else
  position = ((unsigned long)free_area - (unsigned long)alloc_area) / BUFFER_SIZE;
  if ((position < 0) || (position >= MAX_CPU_NUMBER)) goto error;
#endif

#ifdef DEBUG
  printf("  Position : %d\n", position);
#endif  

  alloc_used[position] = 0;
 
#ifdef DEBUG
  printf("Unmap Succeeded.\n\n");
#endif  

  UNLOCK_COMMAND(&alloc_lock);

  return;
  
 error:
  printf("BLAS : Bad memory unallocation! : %4d  %p\n", position,  free_area);

#ifdef DEBUG
  for (position = 0; position < MAX_CPU_NUMBER; position++) 
    printf("%4ld  %p : %d\n", position, alloc_area[position], alloc_used[position]);
#endif

  UNLOCK_COMMAND(&alloc_lock);

  return;
}

==> GotoBLAS-1.10.tar.gz: GotoBLAS/level1/others/memory.c <==
/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#undef  DEBUG
#undef  DEBUG_AFFINITY

#if (!defined(ARCH_X86) && !defined(ARCH_X86_64)) || \
     !defined(SMP_SERVER) || !defined(linux) || (MAX_CPU_NUMBER == 1)
#undef USE_CPU_AFFINITY
#undef USE_MEMORY_AFFINITY
#endif

#if defined(linux) && !defined(USE_CPU_AFFINITY) && defined(USE_MEMORY_AFFINITY)
#define USE_CPU_AFFINITY
#endif

#ifndef ALLOC_STATIC
#if defined( __WIN32__) || defined(__CYGWIN32__)
#define ALLOC_WINDOWS
#else
#define ALLOC_MMAP
#define ALLOC_MALLOC
#endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#ifdef __unix__
#include <pthread.h>
#include <sys/mman.h>
#endif
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#ifdef linux
#include <sys/sysinfo.h>
#include <sched.h>
#include <errno.h>
#include <linux/unistd.h>
#if 0
#ifdef USE_MEMORY_AFFINITY
#include <linux/mempolicy.h>
#endif
#endif
#endif

#if defined( __WIN32__) || defined(__CYGWIN32__)
#include <windows.h>
#endif

#ifdef USE_MEMORY_AFFINITY
#include <numaif.h>
#endif

#include "common.h"

#ifdef ALLOC_HUGETLB
#define SHM_HUGETLB 04000
#endif

#ifdef USE_CPU_AFFINITY

static inline int get_physical_pos(void){
  int eax, ebx, ecx, edx;
  int apicid;

  cpuid(1, &eax, &ebx, &ecx, &edx);
  apicid  = BITMASK(ebx, 24, 0xff);

  return apicid;
}
#endif

#ifdef USE_MEMORY_AFFINITY

/* This function works only on Opteron/Prescott */
static inline int get_num_cores(void){

  int ncores;

#if   defined(CORE_PRESCOTT)
  int eax, ebx, ecx, edx;

  cpuid(4, &eax, &ebx, &ecx, &edx);
  ncores = BITMASK(eax, 26, 0x3f) + 1;
#elif defined(CORE_OPTERON)
  int eax, ebx, ecx, edx;

  cpuid(0x80000008, &eax, &ebx, &ecx, &edx);
  ncores = BITMASK(ecx,  0, 0xff) + 1;
#else
  ncores = 1;
#endif

  return ncores;
}
#endif

#ifndef FIXED_PAGESIZE
#define FIXED_PAGESIZE 4096
#endif

#define BITMASK(a, b, c) ((((a) >> (b)) & (c)))

#ifndef SMP

#define blas_cpu_number 1

#else

/* Global Parameter */
int blas_cpu_number = 0;

int blas_get_cpu_number(void){
  char *p;

  if (blas_cpu_number) return blas_cpu_number;

  p = getenv("GOTO_NUM_THREADS");

  if (p == (char *)NULL) p = getenv("OMP_NUM_THREADS");

  if (p) {
    blas_cpu_number = atoi(p);
#ifdef DEBUG
    printf( "Environment is defined : %3d\n", blas_cpu_number);
#endif
  } else {
#ifdef linux
    blas_cpu_number = get_nprocs();
#endif
  }

  if (blas_cpu_number <= 0) blas_cpu_number = DEFAULT_CPU_NUMBER;

  if (blas_cpu_number > MAX_CPU_NUMBER) blas_cpu_number = MAX_CPU_NUMBER;

#ifdef DEBUG
    printf( "Adjusted number of threads : %3d\n", blas_cpu_number);
#endif

  return blas_cpu_number;
}
#endif

#ifdef ALLOC_STATIC
extern double alloc_area[MAX_CPU_NUMBER * BUFFER_SIZE / sizeof(double)];
#else
#ifdef SEEK_ADDRESS
static unsigned long base_address      = 0UL;
#else
static unsigned long base_address      = BASE_ADDRESS;
#endif
static unsigned long alloc_area[MAX_CPU_NUMBER];
#endif

static int alloc_used[MAX_CPU_NUMBER];

#if   defined(USE_PTHREAD_LOCK)
static pthread_mutex_t    alloc_lock = PTHREAD_MUTEX_INITIALIZER;
#elif defined(USE_PTHREAD_SPINLOCK)
static pthread_spinlock_t alloc_lock = 0;
#else
static unsigned long      alloc_lock = 0UL;
#endif

#ifdef ALLOC_MMAP
static void *alloc_mmap(void *address){
  void *map_address;

  if (address){
    map_address = mmap(address, 
		       BUFFER_SIZE, 
		       MMAP_ACCESS, MMAP_POLICY | MAP_FIXED, -1, 0);
  } else {
    map_address = mmap(address, 
		       BUFFER_SIZE, 
		       MMAP_ACCESS, MMAP_POLICY, -1, 0);
  }

  return map_address;
}
#endif

#ifdef ALLOC_MALLOC
static void *alloc_malloc(void *address){
  void *map_address;

  map_address = (void *)malloc(BUFFER_SIZE + FIXED_PAGESIZE);

  /* Dangerous. Never release this area */
  return (void *)(((long)map_address + FIXED_PAGESIZE - 1) & ~(FIXED_PAGESIZE - 1));
}
#endif

#ifdef ALLOC_QALLOC
void *qalloc(int flags, size_t bytes);
#define QNONCACHE 0x1
#define QCOMMS    0x2
#define QFAST     0x4

static void *alloc_qalloc(void *address){
  void *map_address;

  map_address = (void *)qalloc(QCOMMS | QFAST, BUFFER_SIZE + FIXED_PAGESIZE);

  return (void *)(((long)map_address + FIXED_PAGESIZE - 1) & ~(FIXED_PAGESIZE - 1));
}
#endif

#ifdef ALLOC_WINDOWS
static void *alloc_windows(void *address){
  void *map_address;

  map_address  = VirtualAlloc(address,
			      BUFFER_SIZE,
			      MEM_RESERVE | MEM_COMMIT,
			      PAGE_READWRITE);
  return map_address;
}
#endif

#ifdef ALLOC_SHM
static void *alloc_shm(void *address){
  unsigned long map_address;
  int shmid;
  
  shmid = shmget(IPC_PRIVATE, BUFFER_SIZE,IPC_CREAT | 0600);
  
  map_address = (unsigned long)shmat(shmid, address, 0);

  if ((long)map_address != -1){
    shmctl(shmid, IPC_RMID, 0);
  }

  return (void *)map_address;
}
#endif

#ifdef ALLOC_HUGETLB
static void *alloc_hugetlb(void *address){
  unsigned long map_address = (unsigned long)-1;

#if defined(linux) || defined(_AIX)
  int shmid;
  
  shmid = shmget(IPC_PRIVATE, BUFFER_SIZE,
#ifdef linux
		 SHM_HUGETLB |
#endif
#ifdef _AIX
		 SHM_LGPAGE | SHM_PIN |
#endif
		 IPC_CREAT | SHM_R | SHM_W);
  
  if (shmid != -1) {
    map_address = (unsigned long)shmat(shmid, address, SHM_RND);
    
    if ((long)map_address != -1){
      shmctl(shmid, IPC_RMID, 0);
    }
  }
#endif

#ifdef __sparc__
  struct memcntl_mha mha;
  
  mha.mha_cmd = MHA_MAPSIZE_BSSBRK;
  mha.mha_flags = 0;
  mha.mha_pagesize = HUGE_PAGESIZE;
  memcntl(NULL, 0, MC_HAT_ADVISE, (char *)&mha, 0, 0);

  map_address = (unsigned long)memalign(HUGE_PAGESIZE, BUFFER_SIZE);
#endif

  return (void *)map_address;
}
#endif

#if defined(ARCH_X86) || defined(ARCH_X86_64) || defined(ARCH_IA64)
static int              initialized = 0;
#endif

void *blas_memory_alloc(int numproc){
  
  int position;
#ifndef ALLOC_STATIC
  unsigned long map_address;
#endif
#ifdef USE_CPU_AFFINITY
  unsigned long rv;
  cpu_set_t cpu_mask;
#endif
#ifdef USE_MEMORY_AFFINITY
  unsigned long mem_mask[2];
#endif

  void *(*memoryalloc[])(void *address) = {
#ifdef ALLOC_HUGETLB
    alloc_hugetlb,
#endif
#ifdef ALLOC_SHM
    alloc_shm,
#endif
#ifdef ALLOC_MMAP
    alloc_mmap,
#endif
#ifdef ALLOC_QALLOC
    alloc_qalloc,
#endif
#ifdef ALLOC_WINDOWS
    alloc_windows,
#endif
#ifdef ALLOC_MALLOC
    alloc_malloc,
#endif
    NULL,
  };
  void *(**func)(void *address);

#ifdef SMP
  if (!blas_cpu_number) blas_get_cpu_number();
#endif

  LOCK_COMMAND(&alloc_lock);
  
#ifdef DEBUG
  printf("Alloc Start ...\n");
#endif

  /* main thread allocate to main CPU */
#if MAX_CPU_NUMBER == 1
  position = 0;
#else
  if (numproc >= 2) position = 1; else position = 0;
#endif

#ifdef USE_CPU_AFFINITY
  position = get_physical_pos();
  if ((position >= MAX_CPU_NUMBER) || alloc_used[position])
    position = MAX_CPU_NUMBER - 1;
  while (alloc_used[position]) position--;
#else
  while (alloc_used[position]) position++;
#endif

  if ((position < 0) || (position >= MAX_CPU_NUMBER)) goto error;

#ifdef DEBUG
  printf("  Position -> %d\n", position);
#endif

#ifndef ALLOC_STATIC

  if (!alloc_area[position]) {
    do {
#ifdef DEBUG
      printf("Allocation Start : %lx\n", base_address);
#endif

#ifdef USE_CPU_AFFINITY
      rv = sched_getaffinity((pid_t)0, sizeof(cpu_mask), &cpu_mask);
      if ((rv == 0) && CPU_ISSET(position, &cpu_mask)) {
	CPU_ZERO(&cpu_mask);
	CPU_SET(position, &cpu_mask);
	rv = sched_setaffinity((pid_t)0, sizeof(cpu_mask), &cpu_mask);
      }
#endif

      map_address = (unsigned long)-1;
      func = &memoryalloc[0];

      while ((func != NULL) && (map_address == (unsigned long) -1)) {
	map_address = (unsigned long)(*func)((void *)base_address);
	func ++;
      }

#ifdef USE_MEMORY_AFFINITY
      if (map_address) {
	mem_mask[0] = (1 << (position >> get_num_cores()));
	mem_mask[1] = 0;
	rv = mbind((void *)map_address, BUFFER_SIZE, 
		   MPOL_BIND, mem_mask, sizeof(mem_mask), MPOL_MF_STRICT);
      }
#endif

#ifdef DEBUG
      printf("  Success -> %08lx\n", map_address);
#endif
      if (((long) map_address) == -1) base_address = 0UL;
      if (base_address) base_address += BUFFER_SIZE;

    } while ((long)map_address == -1);

    alloc_area[position] = map_address;

#ifdef DEBUG
    printf("  Mapping Succeeded. %p(%d)\n", (void *)alloc_area[position], position);
#endif
  }
 
#endif

  alloc_used[position] = 1;

#ifdef DEBUG
  printf("Mapped   : %p  %3d\n\n",
	  (void *)alloc_area[position], position);
#endif  

#if defined(ARCH_X86) || defined(ARCH_X86_64) || defined(ARCH_IA64)
  if (initialized == 0) {
    blas_set_parameter();
    initialized = 1;
  }
#endif

  UNLOCK_COMMAND(&alloc_lock);
  
#ifndef ALLOC_STATIC
  return (void *)alloc_area[position];
#else
  return ((void *)alloc_area) + position * BUFFER_SIZE;
#endif

 error:
  UNLOCK_COMMAND(&alloc_lock);
  printf("BLAS : Program is Terminated. Because you tried to allocate too many memory regions.\n");
  printf("       If you use OpenMP, MAX_THREADS in Makefile.rule should be larger than\n");
  printf("       OMP_NUM_THREADS * GOTO_NUM_THREADS.\n");
  exit(1);
}

void blas_memory_free(void *free_area){

  int position;

  LOCK_COMMAND(&alloc_lock);

#ifdef DEBUG
  printf("Unmapped Start : %p ...\n", free_area);
#endif  

#ifndef ALLOC_STATIC
  position = 0;

  while (((unsigned long)alloc_area[position] != (unsigned long)free_area)
	 && (position < MAX_CPU_NUMBER)) position++;

  if ((unsigned long)alloc_area[position] != (unsigned long)free_area) goto error;
#else
  position = ((unsigned long)free_area - (unsigned long)alloc_area) / BUFFER_SIZE;
  if ((position < 0) || (position >= MAX_CPU_NUMBER)) goto error;
#endif

#ifdef DEBUG
  printf("  Position : %d\n", position);
#endif  

  alloc_used[position] = 0;
 
#ifdef DEBUG
  printf("Unmap Succeeded.\n\n");
#endif  

  UNLOCK_COMMAND(&alloc_lock);

  return;
  
 error:
  printf("BLAS : Bad memory unallocation! : %4d  %p\n", position,  free_area);

#ifdef DEBUG
  for (position = 0; position < MAX_CPU_NUMBER; position++) 
    printf("%4ld  %p : %d\n", position, alloc_area[position], alloc_used[position]);
#endif

  UNLOCK_COMMAND(&alloc_lock);

  return;
}
